@@ -61,30 +61,31 @@ class _LazyModule(_ModuleType):
 
 os = _LazyModule.load('os', 'os')
 shutil = _LazyModule.load('shutil', 'shutil')
-builtins = _LazyModule.load('builtins', 'builtins')
 xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
 xp = _LazyModule.load('xonsh', 'xonsh.platform', 'xp')
+from xonsh.built_ins import XSH
+
+
 def _replace_home(x):
     if xp.ON_WINDOWS:
-        home = (
-            builtins.__xonsh__.env["HOMEDRIVE"] + builtins.__xonsh__.env["HOMEPATH"][0]
-        )
+        home = XSH.env["HOMEDRIVE"] + XSH.env["HOMEPATH"][0]
         if x.startswith(home):
             x = x.replace(home, "~", 1)
 
-        if builtins.__xonsh__.env.get("FORCE_POSIX_PATHS"):
+        if XSH.env.get("FORCE_POSIX_PATHS"):
             x = x.replace(os.sep, os.altsep)
 
         return x
     else:
-        home = builtins.__xonsh__.env["HOME"]
+        home = XSH.env["HOME"]
         if x.startswith(home):
             x = x.replace(home, "~", 1)
         return x
 
 
 def _replace_home_cwd():
-    return _replace_home(builtins.__xonsh__.env["PWD"])
+    pwd = XSH.env["PWD"].replace("{", "{{").replace("}", "}}")
+    return _replace_home(pwd)
 
 
 def _collapsed_pwd():
@@ -105,8 +106,8 @@ def _dynamically_collapsed_pwd():
     environment variable DYNAMIC_CWD_WIDTH.
     """
     original_path = _replace_home_cwd()
-    target_width, units = builtins.__xonsh__.env["DYNAMIC_CWD_WIDTH"]
-    elision_char = builtins.__xonsh__.env["DYNAMIC_CWD_ELISION_CHAR"]
+    target_width, units = XSH.env["DYNAMIC_CWD_WIDTH"]
+    elision_char = XSH.env["DYNAMIC_CWD_ELISION_CHAR"]
     if target_width == float("inf"):
         return original_path
     if units == "%":
@@ -152,15 +153,15 @@ def _dynamically_collapsed_pwd():
 """Prompt formatter for virtualenv and others"""
 
 # amalgamated os
-# amalgamated builtins
+# amalgamated from xonsh.built_ins import XSH
 # amalgamated xonsh.platform
 def find_env_name():
     """Finds the current environment name from $VIRTUAL_ENV or
     $CONDA_DEFAULT_ENV if that is set.
     """
-    env_path = builtins.__xonsh__.env.get("VIRTUAL_ENV", "")
+    env_path = XSH.env.get("VIRTUAL_ENV", "")
     if len(env_path) == 0 and xp.ON_ANACONDA:
-        env_path = builtins.__xonsh__.env.get("CONDA_DEFAULT_ENV", "")
+        env_path = XSH.env.get("CONDA_DEFAULT_ENV", "")
     env_name = os.path.basename(env_path)
     return env_name
 
@@ -170,15 +171,15 @@ def env_name():
     ``{env_prefix}`` and ``{env_postfix}`` fields.
     """
     env_name = find_env_name()
-    if builtins.__xonsh__.env.get("VIRTUAL_ENV_DISABLE_PROMPT") or not env_name:
+    if XSH.env.get("VIRTUAL_ENV_DISABLE_PROMPT") or not env_name:
         # env name prompt printing disabled, or no environment; just return
         return
 
-    venv_prompt = builtins.__xonsh__.env.get("VIRTUAL_ENV_PROMPT")
+    venv_prompt = XSH.env.get("VIRTUAL_ENV_PROMPT")
     if venv_prompt is not None:
         return venv_prompt
     else:
-        pf = builtins.__xonsh__.shell.prompt_formatter
+        pf = XSH.shell.prompt_formatter
         pre = pf._get_field_value("env_prefix")
         post = pf._get_field_value("env_postfix")
         return pre + env_name + post
@@ -191,7 +192,7 @@ def vte_new_tab_cwd():
     on startup. Note that this does not return a string, it simply prints
     and flushes the escape sequence to stdout directly.
     """
-    env = builtins.__xonsh__.env
+    env = XSH.env
     t = "\033]7;file://{}{}\007"
     s = t.format(env.get("HOSTNAME"), env.get("PWD"))
     print(s, end="", flush=True)
@@ -201,39 +202,26 @@ def vte_new_tab_cwd():
 #
 # -*- coding: utf-8 -*-
 """Informative git status prompt formatter"""
-
-# amalgamated builtins
-collections = _LazyModule.load('collections', 'collections')
 # amalgamated os
 subprocess = _LazyModule.load('subprocess', 'subprocess')
-xl = _LazyModule.load('xonsh', 'xonsh.lazyasd', 'xl')
-GitStatus = collections.namedtuple(
-    "GitStatus",
-    [
-        "branch",
-        "num_ahead",
-        "num_behind",
-        "untracked",
-        "changed",
-        "deleted",
-        "conflicts",
-        "staged",
-        "stashed",
-        "operations",
-    ],
-)
+# amalgamated from xonsh.built_ins import XSH
+from xonsh.color_tools import COLORS
+from xonsh.tools import NamedConstantMeta, XAttr
 
 
-def _check_output(*args, **kwargs):
+def _check_output(*args, **kwargs) -> str:
+    denv = XSH.env.detype()
+    denv.update({"GIT_OPTIONAL_LOCKS": "0"})
+
     kwargs.update(
         dict(
-            env=builtins.__xonsh__.env.detype(),
+            env=denv,
             stdout=subprocess.PIPE,
             stderr=subprocess.DEVNULL,
             universal_newlines=True,
         )
     )
-    timeout = builtins.__xonsh__.env["VC_BRANCH_TIMEOUT"]
+    timeout = XSH.env["VC_BRANCH_TIMEOUT"]
     # See https://docs.python.org/3/library/subprocess.html#subprocess.Popen.communicate
     with subprocess.Popen(*args, **kwargs) as proc:
         try:
@@ -260,35 +248,65 @@ def _check_output(*args, **kwargs):
             raise
 
 
-@xl.lazyobject
-def _DEFS():
-    DEFS = {
-        "HASH": ":",
-        "BRANCH": "{CYAN}",
-        "OPERATION": "{CYAN}",
-        "STAGED": "{RED}●",
-        "CONFLICTS": "{RED}×",
-        "CHANGED": "{BLUE}+",
-        "DELETED": "{RED}-",
-        "UNTRACKED": "…",
-        "STASHED": "⚑",
-        "CLEAN": "{BOLD_GREEN}✓",
-        "AHEAD": "↑·",
-        "BEHIND": "↓·",
-    }
-    return DEFS
+class _DEFS(metaclass=NamedConstantMeta):
+    HASH_INDICATOR = XAttr(":")
+    BRANCH = XAttr("{CYAN}")
+    OPERATION = XAttr("{CYAN}")
+    STAGED = XAttr("{RED}●")
+    CONFLICTS = XAttr("{RED}×")
+    CHANGED = XAttr("{BLUE}+")
+    DELETED = XAttr("{RED}-")
+    UNTRACKED = XAttr("…")
+    STASHED = XAttr("⚑")
+    CLEAN = XAttr("{BOLD_GREEN}✓")
+    AHEAD = XAttr("↑·")
+    BEHIND = XAttr("↓·")
+    LINES_ADDED = XAttr("{BLUE}+")
+    LINES_REMOVED = XAttr("{RED}-")
+    SEPARATOR = XAttr("{RESET}|")
 
 
-def _get_def(key):
-    def_ = builtins.__xonsh__.env.get("XONSH_GITSTATUS_" + key)
-    return def_ if def_ is not None else _DEFS[key]
+def _get_def(attr: XAttr) -> str:
+    key = attr.name.upper()
+    def_ = XSH.env.get("XONSH_GITSTATUS_" + key)
+    return def_ if def_ is not None else attr.value
 
 
-def _get_tag_or_hash():
+def _is_hidden(attr: XAttr) -> bool:
+    hidden = XSH.env.get("XONSH_GITSTATUS_FIELDS_HIDDEN") or set()
+    return attr.name.upper() in hidden or attr.name.lower() in hidden
+
+
+def _get_tag_or_hash() -> str:
     tag_or_hash = _check_output(["git", "describe", "--always"]).strip()
     hash_ = _check_output(["git", "rev-parse", "--short", "HEAD"]).strip()
     have_tag_name = tag_or_hash != hash_
-    return tag_or_hash if have_tag_name else _get_def("HASH") + hash_
+    return tag_or_hash if have_tag_name else _get_def(_DEFS.HASH_INDICATOR) + hash_
+
+
+def _parse_int(val: str, default=0):
+    if val.isdigit():
+        return int(val)
+    return default
+
+
+def _get_files_changed():
+    try:
+        changed = _check_output(["git", "diff", "--numstat"])
+    except subprocess.CalledProcessError:
+        return {}
+
+    insert = 0
+    delete = 0
+
+    if changed:
+        for line in changed.splitlines():
+            x = line.split(maxsplit=2)
+            if len(x) > 1:
+                insert += _parse_int(x[0])
+                delete += _parse_int(x[1])
+
+    return {_DEFS.LINES_ADDED: insert, _DEFS.LINES_REMOVED: delete}
 
 
 def _get_stash(gitdir):
@@ -308,17 +326,25 @@ def _gitoperation(gitdir):
         ("REVERT_HEAD", "REVERTING"),
         ("BISECT_LOG", "BISECTING"),
     )
-    return [f[1] for f in files if os.path.exists(os.path.join(gitdir, f[0]))]
+    operations = [f[1] for f in files if os.path.exists(os.path.join(gitdir, f[0]))]
+    if operations:
+        return "|" + "|".join(operations)
+    return ""
 
 
-def gitstatus():
-    """Return namedtuple with fields:
-    branch name, number of ahead commit, number of behind commit,
-    untracked number, changed number, deleted number, conflicts number,
-    staged number, stashed number, operation."""
+def _get_operation_stashed():
+    gitdir = _check_output(["git", "rev-parse", "--git-dir"]).strip()
+    stashed = _get_stash(gitdir)
+    operation = _gitoperation(gitdir)
+    return {_DEFS.OPERATION: operation, _DEFS.STASHED: stashed}
+
+
+def _get_status_fields():
+    """Return parsed values from ``git status``"""
+
     status = _check_output(["git", "status", "--porcelain", "--branch"])
     branch = ""
-    num_ahead, num_behind = 0, 0
+    ahead, behind = 0, 0
     untracked, changed, deleted, conflicts, staged = 0, 0, 0, 0, 0
     for line in status.splitlines():
         if line.startswith("##"):
@@ -336,9 +362,9 @@ def gitstatus():
                     divergence = divergence.strip("[]")
                     for div in divergence.split(", "):
                         if "ahead" in div:
-                            num_ahead = int(div[len("ahead ") :].strip())
+                            ahead = int(div[len("ahead ") :].strip())
                         elif "behind" in div:
-                            num_behind = int(div[len("behind ") :].strip())
+                            behind = int(div[len("behind ") :].strip())
         elif line.startswith("??"):
             untracked += 1
         else:
@@ -352,57 +378,104 @@ def gitstatus():
             elif len(line) > 0 and line[0] != " ":
                 staged += 1
 
-    gitdir = _check_output(["git", "rev-parse", "--git-dir"]).strip()
-    stashed = _get_stash(gitdir)
-    operations = _gitoperation(gitdir)
+    return {
+        _DEFS.BRANCH: branch,
+        _DEFS.AHEAD: ahead,
+        _DEFS.BEHIND: behind,
+        _DEFS.UNTRACKED: untracked,
+        _DEFS.CHANGED: changed,
+        _DEFS.DELETED: deleted,
+        _DEFS.CONFLICTS: conflicts,
+        _DEFS.STAGED: staged,
+    }
 
-    return GitStatus(
-        branch,
-        num_ahead,
-        num_behind,
-        untracked,
-        changed,
-        deleted,
-        conflicts,
-        staged,
-        stashed,
-        operations,
-    )
+
+def get_gitstatus_fields():
+    """return all shown fields"""
+    fields = {}
+    for keys, provider in (
+        (
+            (
+                _DEFS.BRANCH,
+                _DEFS.AHEAD,
+                _DEFS.BEHIND,
+                _DEFS.UNTRACKED,
+                _DEFS.CHANGED,
+                _DEFS.DELETED,
+                _DEFS.CONFLICTS,
+                _DEFS.STAGED,
+            ),
+            _get_status_fields,
+        ),
+        (
+            (
+                _DEFS.OPERATION,
+                _DEFS.STASHED,
+            ),
+            _get_operation_stashed,
+        ),
+        (
+            (
+                _DEFS.LINES_ADDED,
+                _DEFS.LINES_REMOVED,
+            ),
+            _get_files_changed,
+        ),
+    ):
+        if any(map(lambda x: not _is_hidden(x), keys)):
+            try:
+                fields.update(provider())
+            except subprocess.SubprocessError:
+                return None
+    return fields
 
 
 def gitstatus_prompt():
     """Return str `BRANCH|OPERATOR|numbers`"""
-    try:
-        s = gitstatus()
-    except subprocess.SubprocessError:
+    fields = get_gitstatus_fields()
+    if fields is None:
         return None
 
-    ret = _get_def("BRANCH") + s.branch
-    if s.num_ahead > 0:
-        ret += _get_def("AHEAD") + str(s.num_ahead)
-    if s.num_behind > 0:
-        ret += _get_def("BEHIND") + str(s.num_behind)
-    if s.operations:
-        ret += _get_def("OPERATION") + "|" + "|".join(s.operations)
-    ret += "|"
-    for category in (
-        "staged",
-        "conflicts",
-        "changed",
-        "deleted",
-        "untracked",
-        "stashed",
-    ):
-        symbol = _get_def(category.upper())
-        value = getattr(s, category)
+    ret = ""
+    for fld in (_DEFS.BRANCH, _DEFS.AHEAD, _DEFS.BEHIND, _DEFS.OPERATION):
+        if not _is_hidden(fld):
+            val = fields[fld]
+            if not val:
+                continue
+            ret += _get_def(fld) + str(val)
+
+    if ret:
+        ret += "|"
+
+    number_flds = (
+        _DEFS.STAGED,
+        _DEFS.CONFLICTS,
+        _DEFS.CHANGED,
+        _DEFS.DELETED,
+        _DEFS.UNTRACKED,
+        _DEFS.STASHED,
+        _DEFS.LINES_ADDED,
+        _DEFS.LINES_REMOVED,
+    )
+    for fld in number_flds:
+        if _is_hidden(fld):
+            continue
+        symbol = _get_def(fld)
+        value = fields[fld]
         if symbol and value > 0:
-            ret += symbol + str(value) + "{RESET}"
-    if s.staged + s.conflicts + s.changed + s.deleted + s.untracked + s.stashed == 0:
-        symbol = _get_def("CLEAN")
+            ret += symbol + str(value) + COLORS.RESET
+    if sum((fields.get(fld, 0) for fld in number_flds)) == 0 and not _is_hidden(
+        _DEFS.CLEAN
+    ):
+        symbol = _get_def(_DEFS.CLEAN)
         if symbol:
-            ret += symbol + "{RESET}"
+            ret += symbol + COLORS.RESET
     ret = ret.rstrip("|")
-    ret += "{RESET}"
+
+    if not ret.endswith(COLORS.RESET):
+        ret += COLORS.RESET
+
+    ret = ret.replace("|", _get_def(_DEFS.SEPARATOR))
 
     return ret
 
@@ -429,9 +502,9 @@ def _current_job():
 # -*- coding: utf-8 -*-
 """date & time related prompt formatter"""
 time = _LazyModule.load('time', 'time')
-# amalgamated builtins
+# amalgamated from xonsh.built_ins import XSH
 def _localtime():
-    pf = builtins.__xonsh__.env.get("PROMPT_FIELDS", {})
+    pf = XSH.env.get("PROMPT_FIELDS", {})
     tf = pf.get("time_format", "%H:%M:%S")
     return time.strftime(tf, time.localtime())
 
@@ -443,14 +516,14 @@ def _localtime():
 # pylint:disable=no-member, invalid-name
 
 # amalgamated os
-sys = _LazyModule.load('sys', 'sys')
-queue = _LazyModule.load('queue', 'queue')
-# amalgamated builtins
-threading = _LazyModule.load('threading', 'threading')
-# amalgamated subprocess
-re = _LazyModule.load('re', 're')
 pathlib = _LazyModule.load('pathlib', 'pathlib')
+queue = _LazyModule.load('queue', 'queue')
+re = _LazyModule.load('re', 're')
+# amalgamated subprocess
+sys = _LazyModule.load('sys', 'sys')
+threading = _LazyModule.load('threading', 'threading')
 # amalgamated xonsh.tools
+# amalgamated from xonsh.built_ins import XSH
 from xonsh.lazyasd import LazyObject
 
 RE_REMOVE_ANSI = LazyObject(
@@ -463,7 +536,7 @@ RE_REMOVE_ANSI = LazyObject(
 def _run_git_cmd(cmd):
     # create a safe detyped env dictionary and update with the additional git environment variables
     # when running git status commands we do not want to acquire locks running command like git status
-    denv = dict(builtins.__xonsh__.env.detype())
+    denv = dict(XSH.env.detype())
     denv.update({"GIT_OPTIONAL_LOCKS": "0"})
     return subprocess.check_output(cmd, env=denv, stderr=subprocess.DEVNULL)
 
@@ -472,11 +545,10 @@ def _get_git_branch(q):
     try:
         cmd = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
         branch = xt.decode_bytes(_run_git_cmd(cmd))
-        branch = branch.splitlines()[0] or None
-    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
+    except (subprocess.CalledProcessError, OSError):
         q.put(None)
     else:
-        q.put(branch)
+        q.put(branch.splitlines()[0] if branch else None)
 
 
 def get_git_branch():
@@ -484,7 +556,7 @@ def get_git_branch():
     be determined (timeout, not in a git repo, etc.) then this returns None.
     """
     branch = None
-    timeout = builtins.__xonsh__.env.get("VC_BRANCH_TIMEOUT")
+    timeout = XSH.env.get("VC_BRANCH_TIMEOUT")
     q = queue.Queue()
 
     t = threading.Thread(target=_get_git_branch, args=(q,))
@@ -500,7 +572,7 @@ def get_git_branch():
 
 
 def _get_hg_root(q):
-    _curpwd = builtins.__xonsh__.env["PWD"]
+    _curpwd = XSH.env["PWD"]
     while True:
         if not os.path.isdir(_curpwd):
             return False
@@ -522,7 +594,7 @@ def get_hg_branch(root=None):
     """Try to get the mercurial branch of the current directory,
     return None if not in a repo or subprocess.TimeoutExpired if timed out.
     """
-    env = builtins.__xonsh__.env
+    env = XSH.env
     timeout = env["VC_BRANCH_TIMEOUT"]
     q = queue.Queue()
     t = threading.Thread(target=_get_hg_root, args=(q,))
@@ -565,7 +637,7 @@ _FIRST_BRANCH_TIMEOUT = True
 
 def _first_branch_timeout_message():
     global _FIRST_BRANCH_TIMEOUT
-    sbtm = builtins.__xonsh__.env["SUPPRESS_BRANCH_TIMEOUT_MESSAGE"]
+    sbtm = XSH.env["SUPPRESS_BRANCH_TIMEOUT_MESSAGE"]
     if not _FIRST_BRANCH_TIMEOUT or sbtm:
         return
     _FIRST_BRANCH_TIMEOUT = False
@@ -582,8 +654,8 @@ def _first_branch_timeout_message():
 
 
 def _vc_has(binary):
-    """ This allows us to locate binaries after git only if necessary """
-    cmds = builtins.__xonsh__.commands_cache
+    """This allows us to locate binaries after git only if necessary"""
+    cmds = XSH.commands_cache
     if cmds.is_empty():
         return bool(cmds.locate_binary(binary, ignore_alias=True))
     else:
@@ -619,7 +691,7 @@ def _git_dirty_working_directory(q, include_untracked):
             q.put(bool(status))
         else:
             q.put(None)
-    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
+    except (subprocess.CalledProcessError, OSError):
         q.put(None)
 
 
@@ -627,7 +699,7 @@ def git_dirty_working_directory():
     """Returns whether or not the git directory is dirty. If this could not
     be determined (timeout, file not found, etc.) then this returns None.
     """
-    env = builtins.__xonsh__.env
+    env = XSH.env
     timeout = env.get("VC_BRANCH_TIMEOUT")
     include_untracked = env.get("VC_GIT_INCLUDE_UNTRACKED")
     q = queue.Queue()
@@ -646,7 +718,7 @@ def hg_dirty_working_directory():
     """Computes whether or not the mercurial working directory is dirty or not.
     If this cannot be determined, None is returned.
     """
-    env = builtins.__xonsh__.env
+    env = XSH.env
     cwd = env["PWD"]
     denv = env.detype()
     vcbt = env["VC_BRANCH_TIMEOUT"]
@@ -717,16 +789,16 @@ def branch_bg_color():
 # -*- coding: utf-8 -*-
 """Base prompt, provides PROMPT_FIELDS and prompt related functions"""
 
-# amalgamated builtins
 itertools = _LazyModule.load('itertools', 'itertools')
 # amalgamated os
 # amalgamated re
 socket = _LazyModule.load('socket', 'socket')
 # amalgamated sys
 tp = _LazyModule.load('typing', 'typing', 'tp')
-# amalgamated xonsh.lazyasd
+xl = _LazyModule.load('xonsh', 'xonsh.lazyasd', 'xl')
 # amalgamated xonsh.tools
 # amalgamated xonsh.platform
+# amalgamated from xonsh.built_ins import XSH
 # amalgamated xonsh.prompt.cwd
 # amalgamated xonsh.prompt.job
 # amalgamated xonsh.prompt.env
@@ -751,7 +823,7 @@ class ParsedTokens(tp.NamedTuple):
 
     def process(self) -> str:
         """Wrapper that gets formatter-function from environment and returns final prompt."""
-        processor = builtins.__xonsh__.env.get(  # type: ignore
+        processor = XSH.env.get(  # type: ignore
             "PROMPT_TOKENS_FORMATTER", prompt_tokens_formatter_default
         )
         return processor(self)
@@ -801,7 +873,7 @@ class PromptFormatter:
         self.cache.clear()
 
         if fields is None:
-            self.fields = builtins.__xonsh__.env.get("PROMPT_FIELDS", PROMPT_FIELDS)  # type: ignore
+            self.fields = XSH.env.get("PROMPT_FIELDS", PROMPT_FIELDS)  # type: ignore
         else:
             self.fields = fields
         try:
@@ -832,7 +904,7 @@ class PromptFormatter:
         if field is None:
             return
         elif field.startswith("$"):
-            val = builtins.__xonsh__.env[field[1:]]
+            val = XSH.env[field[1:]]
             return _format_value(val, spec, conv)
         elif field in self.fields:
             val = self._get_field_value(field, spec=spec, conv=conv, **kwargs)
@@ -934,7 +1006,7 @@ def multiline_prompt(curr=""):
     # tail is the trailing whitespace
     tail = line if headlen == 0 else line.rsplit(head[-1], 1)[1]
     # now to construct the actual string
-    dots = builtins.__xonsh__.env.get("MULTILINE_PROMPT")
+    dots = XSH.env.get("MULTILINE_PROMPT")
     dots = dots() if callable(dots) else dots
     if dots is None or len(dots) == 0:
         return ""
@@ -980,7 +1052,7 @@ def is_template_string(template, PROMPT_FIELDS=None):
         return False
     included_names.discard(None)
     if PROMPT_FIELDS is None:
-        fmtter = builtins.__xonsh__.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
+        fmtter = XSH.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
     else:
         fmtter = PROMPT_FIELDS
     known_names = set(fmtter.keys())
